# Copyright 2023 Goldman Sachs
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Callable as PyLegendCallable
from typing import Dict as PyLegendDict
from typing import Iterator as PyLegendIterator
from typing import List as PyLegendList
from typing import Optional as PyLegendOptional
from typing import Sequence as PyLegendSequence
from typing import Type as PyLegendType
from typing import TypeVar as PyLegendTypeVar
from typing import Tuple as PyLegendTuple
from typing import Generator as PyLegendGenerator
from typing import Union as PyLegendUnion
from typing import Generic as PyLegendGeneric
from typing import TYPE_CHECKING


__all__: PyLegendSequence[str] = [
    "PyLegendCallable",
    "PyLegendDict",
    "PyLegendIterator",
    "PyLegendList",
    "PyLegendSequence",
    "PyLegendType",
    "PyLegendTypeVar",
    "PyLegendOptional",
    "PyLegendTuple",
    "PyLegendGenerator",
    "PyLegendUnion",
    "PyLegendGeneric",
    "TYPE_CHECKING",
]
